/*
 * This is an adapted version of the clock.c in the util-linux-2.2
 * distribution.  I modified it to make it work with Linux/Alpha.
 * It should be reasonably straight-forward to unify this source
 * with the x86 version but I haven't had much luck in contacting
 * the util-linux maintainers these days.
 *
 * The only major change was due to the fact that Linux/Alpha cannot
 * disable interrupts at the user-level (see more on this below).
 * Also, the Alpha firmware sets up the CMOS clock in binary, not in
 * BCD mode.  Other than that, pretty much all remained the same.
 *
 * Fri Aug  4 20:06:19 1995		davidm@azstarnet.com
 */
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>

/* V1.0
 * CMOS clock manipulation - Charles Hedrick, hedrick@cs.rutgers.edu, Apr 1992
 * 
 * clock [-u] -r  - read cmos clock
 * clock [-u] -w  - write cmos clock from system time
 * clock [-u] -s  - set system time from cmos clock
 * clock [-u] -a  - set system time from cmos clock, adjust the time to
 *                  correct for systematic error, and put it back to the cmos.
 *  -u indicates cmos clock is kept in universal time
 *  -A indicates cmos clock is kept in Alpha ARC console time (0 == 1980)
 *  -J indicates we're dealing with a Jensen (early DEC Alpha PC)
 *
 * The program is designed to run setuid, since we need to be able to
 * write the CMOS port.
 *
 * I don't know what the CMOS clock will do in 2000, so this program
 * probably won't work past the century boundary.
 *
 *********************
 * V1.1
 * Modified for clock adjustments - Rob Hooft, hooft@chem.ruu.nl, Nov 1992
 * Also moved error messages to stderr. The program now uses getopt.
 * Changed some exit codes. Made 'gcc 2.3 -Wall' happy.
 *
 * I think a small explanation of the adjustment routine should be given
 * here. The problem with my machine is that its CMOS clock is 10 seconds 
 * per day slow. With this version of clock.c, and my '/etc/rc.local' 
 * reading '/etc/clock -au' instead of '/etc/clock -u -s', this error 
 * is automatically corrected at every boot. 
 *
 * To do this job, the program reads and writes the file '/etc/adjtime' 
 * to determine the correction, and to save its data. In this file are 
 * three numbers: 
 *
 * 1) the correction in seconds per day (So if your clock runs 5 
 *    seconds per day fast, the first number should read -5.0)
 * 2) the number of seconds since 1/1/1970 the last time the program was
 *    used.
 * 3) the remaining part of a second which was leftover after the last 
 *    adjustment
 *
 * Installation and use of this program:
 *
 * a) create a file '/etc/adjtime' containing as the first and only line:
 *    '0.0 0 0.0'
 * b) run 'clock -au' or 'clock -a', depending on whether your cmos is in
 *    universal or local time. This updates the second number.
 * c) set your system time using the 'date' command.
 * d) update your cmos time using 'clock -wu' or 'clock -w'
 * e) replace the first number in /etc/adjtime by your correction.
 * f) put the command 'clock -au' or 'clock -a' in your '/etc/rc.local'
 *
 * If the adjustment doesn't work for you, try contacting me by E-mail.
 *
 ******
 * V1.2
 *
 * Applied patches by Harald Koenig (koenig@nova.tat.physik.uni-tuebingen.de)
 * Patched and indented by Rob Hooft (hooft@EMBL-Heidelberg.DE)
 * 
 * A free quote from a MAIL-message (with spelling corrections):
 *
 * "I found the explanation and solution for the CMOS reading 0xff problem
 *  in the 0.99pl13c (ALPHA) kernel: the RTC goes offline for a small amount
 *  of time for updating. Solution is included in the kernel source 
 *  (linux/kernel/time.c)."
 *
 * "I modified clock.c to fix this problem and added an option (now default,
 *  look for USE_INLINE_ASM_IO) that I/O instructions are used as inline
 *  code and not via /dev/port (still possible via #undef ...)."
 *
 * With the new code, which is partially taken from the kernel sources, 
 * the CMOS clock handling looks much more "official".
 * Thanks Harald (and Torsten for the kernel code)!
 *
 ******
 * V1.3
 * Canges from alan@spri.levels.unisa.edu.au (Alan Modra):
 * a) Fix a few typos in comments and remove reference to making
 *    clock -u a cron job.  The kernel adjusts cmos time every 11
 *    minutes - see kernel/sched.c and kernel/time.c set_rtc_mmss().
 *    This means we should really have a cron job updating
 *    /etc/adjtime every 11 mins (set last_time to the current time
 *    and not_adjusted to ???).
 * b) Swapped arguments of outb() to agree with asm/io.h macro of the
 *    same name.  Use outb() from asm/io.h as it's slightly better.
 * c) Changed CMOS_READ and CMOS_WRITE to inline functions.  Inserted
 *    cli()..sti() pairs in appropriate places to prevent possible
 *    errors, and changed ioperm() call to iopl() to allow cli.
 * d) Moved some variables around to localise them a bit.
 * e) Fixed bug with clock -ua or clock -us that cleared environment
 *    variable TZ.  This fix also cured the annoying display of bogus
 *    day of week on a number of machines. (Use mktime(), ctime()
 *    rather than asctime() )
 * f) Use settimeofday() rather than stime().  This one is important
 *    as it sets the kernel's timezone offset, which is returned by
 *    gettimeofday(), and used for display of MSDOS and OS2 file
 *    times.
 * g) faith@cs.unc.edu added -D flag for debugging
 *
 * V1.4: alan@SPRI.Levels.UniSA.Edu.Au (Alan Modra)
 *       Wed Feb  8 12:29:08 1995, fix for years > 2000.
 *       faith@cs.unc.edu added -v option to print version.
 *
 * V1.4axp: davidm@azstarnet.com (David Mosberger)
 *	 Fri Aug  4 18:53:56 1995, modified for use Linux/Alpha.
 *
 * V1.4.1axp: jestabro@amt.tay1.dec.com (Jay Estabrook)
 *	 Mon Dec  4 22:53:56 1995, modified for Linux/Alpha ARC console.
 *
 * V1.4.2axp: jestabro@amt.tay1.dec.com (Jay Estabrook)
 *	 Wed Jan 10 1996, modified to use /dev/port, not inb/outb.
 *       Reason: LIBC support for inb/outb is hardwired to Alpha EV4 CPU.
 *
 * V1.4.3axp: ost@coments.rwth-aachen.de (Martin Ostermann)
 *	 Wed May 22 1996, modified to support J option for the Jensen
 *       Reason: on a Jensen the RTC is at 0x170 instead of 0x70 
 */

#define VERSION "1.4.3axp"

#undef USE_DEV_PORT

#ifdef USE_DEV_PORT
# include <asm/fcntl.h>
#else
# include <asm/io.h>
#endif

/* Here the information for time adjustments is kept. */
#define ADJPATH "/etc/adjtime"


/* used for debugging the code. */
#undef KEEP_OFF

#define cmos_read(reg)		atomic("clock read", op_read, (reg), 0)
#define cmos_write(reg,val)	atomic("clock write", op_write, (reg), (val))
#define cmos_set_time(tm)	atomic("set time", op_set_time, (unsigned long) tm, 0)


/* Globals */
int readit = 0;
int adjustit = 0;
int writeit = 0;
int setit = 0;
int universal = 0;
int debug = 0;
int arcit = 0;
int clockport = 0x70;
const char *pgm_name;

#ifdef USE_DEV_PORT
int dev_port_fd;
#endif

volatile void 
usage ()
{
  fprintf (stderr, 
    "usage: %s [-u] -r|w|s|a|v\n"
    "  r: read and print CMOS clock\n"
    "  w: write CMOS clock from system time\n"
    "  s: set system time from CMOS clock\n"
    "  a: get system time and adjust CMOS clock\n"
    "  u: CMOS clock is in universal time\n"
    "  A: CMOS clock is in Alpha ARC console time\n"
    "  J: we are running on an Alpha Jensen\n"
    "  v: print version (" VERSION ") and exit\n", pgm_name
  );
  exit (1);
}

#ifdef __alpha__

/*
 * The Alpha doesn't allow user-level code to disable interrupts (for
 * good reasons).  Instead, we ensure atomic operation by performing
 * the operation and checking whether the high 32 bits of the cycle
 * counter changed.  If they did, a context switch must have occurred
 * and we redo the operation.  As long as the operation is reasonably
 * short, it will complete atomically, eventually.
 */

static unsigned long
atomic(const char *name, unsigned long (*op)(unsigned long arg1,
					     unsigned long arg2),
       unsigned long arg1, unsigned long arg2)
{
  unsigned long ts1, ts2, n, v;

  for (n = 0; n < 1000; ++n) {
    asm volatile ("rpcc %0" : "r="(ts1));
    v = (*op)(arg1, arg2);
    asm volatile ("rpcc %0" : "r="(ts2));

    if ((ts1 ^ ts2) >> 32 == 0) {
      return v;
    }
  }
  fprintf(stderr, "%s: atomic %s failed for 1000 iterations!", pgm_name, name);
  exit(1);
}

#elif __i386__

static unsigned long
atomic(const char *name, unsigned long (*op)(unsigned long arg1,
					     unsigned long arg2),
       unsigned long arg1, unsigned long arg2)
{
  __asm__ volatile ("cli");
  v = (*op)(arg1, arg2);
  __asm__ volatile ("sti");
}

#else

/*
 * Hmmh, this isn't very atomic.  Maybe we should force an error
 * instead?
 */
static unsigned long
atomic(const char *name, unsigned long (*op)(unsigned long arg1,
					     unsigned long arg2),
       unsigned long arg1, unsigned long arg2)
{
  v = (*op)(arg1, arg2);
}

#endif


unsigned long op_read(unsigned long reg, unsigned long arg2)
{
  unsigned char v;

#ifdef USE_DEV_PORT
  v = reg | 0x80;
  lseek(dev_port_fd, clockport, 0);
  write(dev_port_fd, &v, 1);
  lseek(dev_port_fd, clockport + 1, 0);
  read(dev_port_fd, &v, 1);
  return v;
#else
  outb (reg | 0x80, clockport);
  return inb (clockport + 1);
#endif
}

unsigned long op_write(unsigned long reg, unsigned long val)
{
  unsigned char v;

#ifdef USE_DEV_PORT
  v = reg | 0x80;
  lseek(dev_port_fd, clockport, 0);
  write(dev_port_fd, &v, 1);
  v = val & 0xff;
  lseek(dev_port_fd, clockport + 1, 0);
  write(dev_port_fd, &v, 1);
#else
  outb (reg | 0x80, clockport);
  outb (val, clockport + 1);
#endif
  return 0;
}

unsigned long op_set_time(unsigned long arg1, unsigned long arg2)
{
  unsigned char save_control, save_freq_select;
  struct tm *tm = (struct tm *) arg1;

  save_control = cmos_read (11);   /* tell the clock it's being set */
  cmos_write (11, (save_control | 0x80));
  save_freq_select = cmos_read (10);       /* stop and reset prescaler */
  cmos_write (10, (save_freq_select | clockport));

  cmos_write (0, tm->tm_sec);
  cmos_write (2, tm->tm_min);
  cmos_write (4, tm->tm_hour);
  cmos_write (6, tm->tm_wday + 3);
  cmos_write (7, tm->tm_mday);
  cmos_write (8, tm->tm_mon + 1);
  if (arcit) /* Alpha ARC console */
    tm->tm_year -= 80;
  cmos_write (9, tm->tm_year);

  cmos_write (10, save_freq_select);
  cmos_write (11, save_control);
}

void 
cmos_init ()
{
#ifdef USE_DEV_PORT
  if ((dev_port_fd = open("/dev/port", O_RDWR)) < 0) {
    perror("Cannot open /dev/port");
    exit(1);
  }
#else
  if (ioperm (clockport, 2, 1))
    {
      fprintf(stderr,"%s: unable to get I/O port access\n", pgm_name);
      exit (1);
    }
#endif
}

int 
main (int argc, char **argv, char **envp)
{
  struct tm tm;
  time_t systime;
  time_t last_time;
  char arg;
  double factor;
  double not_adjusted;
  int adjustment = 0;

  pgm_name = argv[0];
  while ((arg = getopt (argc, argv, "rwsuaAJDv")) != -1)
    {
      switch (arg)
	{
	case 'r':
	  readit = 1;
	  break;
	case 'w':
	  writeit = 1;
	  break;
	case 's':
	  setit = 1;
	  break;
	case 'u':
	  universal = 1;
	  break;
	case 'a':
	  adjustit = 1;
	  break;
        case 'A':
	  arcit = 1;
	  break;
	case 'J':
	  clockport = 0x170;
	  break;
        case 'D':
	  debug = 1;
	  break;
	case 'v':
	  fprintf( stderr, "%s " VERSION "\n" , pgm_name);
	  exit(0);
	default:
	  usage ();
	}
    }

  if (readit + writeit + setit + adjustit > 1)
    usage ();			/* only allow one of these */

  if (!(readit | writeit | setit | adjustit))	/* default to read */
    readit = 1;

#ifdef __alpha__
  /* See whether we're dealing with a Jensen---it has a weird I/O
     system.  DEC was just learning how to build Alpha PCs.  */
  if (clockport != 0x170) {
    FILE *cpuinfo;
    char field[256];

    if (cpuinfo = fopen ("/proc/cpuinfo", "r"))
      {
	while (!feof(cpuinfo))
	  {
	    if (fscanf (cpuinfo,"system type : %255s", field) == 1)
	      {
		if (strcmp(field, "Jensen") == 0)
		  clockport = 0x170;
		break;
	      }
	    fgets (field, 256, cpuinfo);
	  }
	fclose(cpuinfo);
      }
  }
#endif

  cmos_init ();

  if (adjustit)
    {				/* Read adjustment parameters first */
      FILE *adj;
      if ((adj = fopen (ADJPATH, "r")) == NULL)
	{
	  perror (ADJPATH);
	  exit (2);
	}
      if (fscanf (adj, "%lf %d %lf", &factor, &last_time, &not_adjusted) < 0)
	{
	  perror (ADJPATH);
	  exit (2);
	}
      fclose (adj);
      if (debug) printf ("Last adjustment done at %d seconds after 1/1/1970\n", last_time);
    }

  if (readit || setit || adjustit)
    {
      int i;

/* read RTC exactly on falling edge of update flag */
/* Wait for rise.... (may take upto 1 second) */

      for (i = 0; i < 10000000; i++)	
	if (cmos_read (10) & 0x80)
	  break;

/* Wait for fall.... (must try at least 2.228 ms) */

      for (i = 0; i < 1000000; i++)	
	if (!(cmos_read (10) & 0x80))
	  break;

/* The purpose of the "do" loop is called "low-risk programming" */
/* In theory it should never run more than once */
      do
	{ 
	  tm.tm_sec = cmos_read (0);
	  tm.tm_min = cmos_read (2);
	  tm.tm_hour = cmos_read (4);
	  tm.tm_wday = cmos_read (6);
	  tm.tm_mday = cmos_read (7);
	  tm.tm_mon = cmos_read (8);
	  tm.tm_year = cmos_read (9);
	}
      while (tm.tm_sec != cmos_read (0));
      if (arcit) { /* Alpha ARC console */
	tm.tm_year += 80;
      } else {
	if (tm.tm_year < 70)
	  tm.tm_year += 100;  /* 70..99 => 1970..1999, 0..69 => 2000..2069 */
      }
      tm.tm_mon--;		/* DOS uses 1 base */
#ifndef __alpha__
      tm.tm_wday -= 3;		/* DOS uses 3 - 9 for week days */
#endif
      tm.tm_isdst = -1;		/* don't know whether it's daylight */
      if (debug) {
	  printf ("CMOS time : %.3s %d/%d/%d %d:%d:%d\n",
		  "SunMonTueWedThuFriSat" + (tm.tm_wday * 3),
		  tm.tm_year, tm.tm_mon + 1, tm.tm_mday,
		  tm.tm_hour, tm.tm_min, tm.tm_sec);
      }
    }

  if (readit || setit || adjustit)
    {
/*
 * mktime() assumes we're giving it local time.  If the CMOS clock
 * is in GMT, we have to set up TZ so mktime knows it.  tzset() gets
 * called implicitly by the time code, but only the first time.  When
 * changing the environment variable, better call tzset() explicitly.
 */
      if (universal)
	{
	  char *zone;
	  zone = (char *) getenv ("TZ");	/* save original time zone */
	  (void) putenv ("TZ=UTC");
	  tzset ();
	  systime = mktime (&tm);
	  /* now put back the original zone */
	  if (zone)
	    {

             char *zonebuf;
             zonebuf = malloc (strlen (zone) + 4);
             strcpy (zonebuf, "TZ=");
             strcpy (zonebuf+3, zone);
             putenv (zonebuf);
             free (zonebuf);
	    }
	  else
	    {			/* wasn't one, so clear it */
	      putenv ("TZ");
	    }
	  tzset ();
	}
      else
	{
	  systime = mktime (&tm);
	}
      if (debug) printf ("Number of seconds since 1/1/1970 is %d\n", systime);
    }

  if (readit)
    {
      printf ("%s", ctime (&systime ));
    }

  if (setit || adjustit)
    {
      struct timeval tv;
      struct timezone tz;

/* program is designed to run setuid, be secure! */

      if (getuid () != 0)
	{			
	  fprintf (stderr, "%s: sorry, must be root to set or adjust time\n", pgm_name);
	  exit (2);
	}

      if (adjustit)
	{			/* the actual adjustment */
	  double exact_adjustment;

	  exact_adjustment = ((double) (systime - last_time))
	    * factor / (24 * 60 * 60)
	    + not_adjusted;
	  if (exact_adjustment > 0)
	    adjustment = (int) (exact_adjustment + 0.5);
	  else
	    adjustment = (int) (exact_adjustment - 0.5);
	  not_adjusted = exact_adjustment - (double) adjustment;
	  systime += adjustment;
	  if (debug) {
	     printf ("Time since last adjustment is %d seconds\n",
		     (int) (systime - last_time));
	     printf ("Adjusting time by %d seconds\n",
		     adjustment);
	     printf ("remaining adjustment is %.3f seconds\n",
		     not_adjusted);
	  }
	}
#ifndef KEEP_OFF
      tv.tv_sec = systime;
      tv.tv_usec = 0;
      tz.tz_minuteswest = timezone / 60;
      tz.tz_dsttime = daylight;

      if (settimeofday (&tv, &tz) != 0)
        {
	  fprintf (stderr,
		   "%s: unable to set time -- probably you are not root\n", pgm_name);
	  exit (1);
	}
      
      if (debug) {
	 printf( "Called settimeofday:\n" );
	 printf( "\ttv.tv_sec = %ld, tv.tv_usec = %ld\n",
		 tv.tv_sec, tv.tv_usec );
	 printf( "\ttz.tz_minuteswest = %d, tz.tz_dsttime = %d\n",
		 tz.tz_minuteswest, tz.tz_dsttime );
      }
#endif
    }
  
  if (writeit || (adjustit && adjustment != 0))
    {
      struct tm *tmp;
      systime = time (NULL);
      if (universal)
	tmp = gmtime (&systime);
      else
	tmp = localtime (&systime);

#ifndef KEEP_OFF
      cmos_set_time(tmp);
#endif
      if (debug) printf ("Set to : %d:%d:%d\n", tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
    }
  else
    if (debug) printf ("CMOS clock unchanged.\n");
  /* Save data for next 'adjustit' call */
  if (adjustit)
    {
      FILE *adj;
      if ((adj = fopen (ADJPATH, "w")) == NULL)
	{
	  perror (ADJPATH);
	  exit (2);
	}
      fprintf (adj, "%f %d %f\n", factor, systime, not_adjusted);
      fclose (adj);
    }
  exit (0);
}
